/**
 * Copyright (c) 2016-2017  Zubax Robotics  <info@zubax.com>
 */

#include "util.hpp"
#include "bootloader_interface.hpp"
#include <os.hpp>
#include <current_build_info.hpp>

/// Provided by linker; see the linker script
extern std::uint8_t BootloaderSharedStruct[];

namespace bootloader_interface
{
#if defined(RELEASE_BUILD) && RELEASE_BUILD
constexpr bool ReleaseBuild = true;
#elif defined(DEBUG_BUILD) && DEBUG_BUILD
constexpr bool ReleaseBuild = false;
#else
#    error "Cannot determine the type of build"
#endif

/**
 * This is the Kocherga app image descriptor.
 * Details: https://github.com/Zubax/kocherga
 */
static const volatile struct __attribute__((packed))
{
    std::uint8_t signature[8] = {'A', 'P', 'D', 'e', 's', 'c', '0', '0'};

    std::uint64_t image_crc  = 0;
    std::uint32_t image_size = 0;
    std::uint32_t vcs_commit = current_build_info::VCSCommitID;

    std::uint8_t major_version = FW_VERSION_MAJOR;
    std::uint8_t minor_version = FW_VERSION_MINOR;

    std::uint8_t flags = (ReleaseBuild ? 1U : 0U) | (current_build_info::BuildIsDirty ? 2U : 0U);

    std::uint8_t _reserved_a_ = 0;

    std::uint32_t build_timestamp_utc = current_build_info::BuildTimestampUTC;

} _app_descriptor __attribute__((section(".app_descriptor")));
static_assert(sizeof(_app_descriptor) == 32, "AppInfo must be exactly 32 bytes large");

static chibios_rt::Mutex g_mutex;

FirmwareVersion getFirmwareVersion()
{
    FirmwareVersion x;
    x.image_crc64we       = _app_descriptor.image_crc;
    x.vcs_commit          = _app_descriptor.vcs_commit;
    x.build_timestamp_utc = current_build_info::BuildTimestampUTC;
    x.major               = _app_descriptor.major_version;
    x.minor               = _app_descriptor.minor_version;
    x.dirty_build         = current_build_info::BuildIsDirty;
    x.release_build       = ReleaseBuild;

    return x;
}

const char* getBuildEnvironmentDescription()
{
    return current_build_info::BuildEnvironmentDescription;
}

static inline auto makeMarshaller()
{
    // The storage area is defined in the linker script.
    void* const address = static_cast<void*>(&BootloaderSharedStruct[0]);
    DEBUG_LOG("AppShared address: 0x%x\n", unsigned(address));
    return AppSharedMarshaller<AppShared>(address);
}

std::optional<AppShared> readAndInvalidateSharedStruct()
{
    os::MutexLocker locker(g_mutex);
    return makeMarshaller().readAndErase();
}

void writeSharedStruct(const AppShared& shared)
{
    os::MutexLocker locker(g_mutex);
    makeMarshaller().write(shared);
}

}  // namespace bootloader_interface
