/**
 * Copyright (c) 2016-2017  Zubax Robotics  <info@zubax.com>
 */

#include <ch.hpp>
#include <hal.h>
#include <unistd.h>
#include <cstring>
#include <cstdio>
#include <cstdlib>

#include "board/board.hpp"
#include "bootloader_interface/bootloader_interface.hpp"

namespace app
{
namespace
{
constexpr unsigned WatchdogTimeout = 5;

/**
 * This callback is invoked from the local UAVCAN node (from its own thread!) when the said node
 * receives a firmware update request. The objective here is to set up the bootloader and signal
 * rebooting; once the system is rebooted, the bootloader will know what to do.
 */
auto onFirmwareUpdateRequestedFromUAVCAN(
    const uavcan::ReceivedDataStructure<uavcan::protocol::file::BeginFirmwareUpdate::Request>& request)
{
    static bool already_in_progress = false;

    const std::uint8_t source_node_id =
        ((request.source_node_id > 0) && (request.source_node_id <= uavcan::NodeID::Max) &&
         uavcan::NodeID(request.source_node_id).isUnicast())
            ? request.source_node_id
            : request.getSrcNodeID().get();

    g_logger.println("UAVCAN firmware update request from %d, source %d, path '%s'",
                     request.getSrcNodeID().get(),
                     source_node_id,
                     request.image_file_remote_path.path.c_str());

    if (already_in_progress)
    {
        g_logger.println("UAVCAN firmware update is already in progress, rejecting");
        return uavcan::protocol::file::BeginFirmwareUpdate::Response::ERROR_IN_PROGRESS;
    }

    /*
     * Initializing the app shared structure with proper arguments
     */
    bootloader_interface::AppShared shared;
    shared.can_bus_speed            = uavcan_node::getCANBusBitRate();
    shared.uavcan_node_id           = uavcan_node::getNodeID().get();
    shared.uavcan_fw_server_node_id = source_node_id;
    shared.stay_in_bootloader       = true;

    std::strncpy(static_cast<char*>(&shared.uavcan_file_name[0]),  // This is really messy
                 request.image_file_remote_path.path.c_str(),
                 shared.UAVCANFileNameMaxLength);
    shared.uavcan_file_name[shared.UAVCANFileNameMaxLength - 1] = '\0';

    static_assert(request.image_file_remote_path.path.MaxSize < shared.UAVCANFileNameMaxLength, "Err...");

    g_logger.println("Bootloader args: CAN bus bitrate: %u, local node ID: %d",
                     unsigned(shared.can_bus_speed),
                     shared.uavcan_node_id);

    /*
     * Committing everything
     */
    bootloader_interface::writeSharedStruct(shared);

    os::requestReboot();

    already_in_progress = true;

    g_logger.puts("UAVCAN firmware update initiated");
    return uavcan::protocol::file::BeginFirmwareUpdate::Response::ERROR_OK;
}

/**
 * This callback is invoked from the serial interface module (from its own thread!) when it
 * receives a firmware update request. The objective here is to set up the bootloader and signal
 * rebooting; once the system is rebooted, the bootloader will know what to do.
 * The return value is true on success, false if denied.
 */
bool onFirmwareUpdateRequestedFromSerial()
{
    g_logger.println("Serial firmware update request");

    bootloader_interface::AppShared shared;
    shared.can_bus_speed      = uavcan_node::getCANBusBitRate();
    shared.uavcan_node_id     = uavcan_node::getNodeID().get();
    shared.stay_in_bootloader = true;

    bootloader_interface::writeSharedStruct(shared);
    os::requestReboot();

    return true;
}

board::Watchdog init()
{
    board::Watchdog watchdog = board::init(std::chrono::seconds(WatchdogTimeout));
    watchdog.kick();

    board::setRGBLED(board::RGB::Ones());

    const auto fw_version = bootloader_interface::getFirmwareVersion();

    const auto app_shared = bootloader_interface::readAndInvalidateSharedStruct();
    if (app_shared)
    {
        g_logger.println("Bootloader struct values: CAN bitrate: %u, UAVCAN Node ID: %u",
                         unsigned(app_shared->can_bus_speed),
                         app_shared->uavcan_node_id);
    }
    else
    {
        g_logger.puts("Bootloader struct is NOT present");
    }

    uavcan_node::init(app_shared ? app_shared->can_bus_speed : 0,
                      app_shared ? app_shared->uavcan_node_id : 0,
                      {fw_version.major, fw_version.minor},
                      fw_version.image_crc64we,
                      fw_version.vcs_commit,
                      &onFirmwareUpdateRequestedFromUAVCAN,
                      &onRebootRequested);

    watchdog.kick();
    return watchdog;
}

int main()
{
    auto watchdog = app::init();

    while (!os::isRebootRequested())
    {
        watchdog.kick();
        // <...>
    }

    board::setRGBLED(board::RGB::Ones());
    watchdog.kick();
    ::sleep(1);  // Let other threads terminate properly
    board::restart();

    return 0;
}
